/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "baiduspeechengine_p.h"
#include <unistd.h>
#include <curl/websockets.h>
#include <jsoncpp/json/json.h>
#include "servererror.h"
#include "token.h"
#include "util.h"

bool connectWebSocket(CURL *curl, BaiduSpeechEnginePrivate *engine) {
    CURLcode res;
    std::string callurl =
        "wss://vop.baidu.com/realtime_asr?sn=kylin-ai-subsystem";

    if (curl) {
        std::string result;
        curl_easy_setopt(curl, CURLOPT_URL, callurl.c_str());
        curl_easy_setopt(curl, CURLOPT_CONNECT_ONLY, 2L);
        res = curl_easy_perform(curl);
        if (res != CURLE_OK) {
            engine->lastErrorMessage_ = curl_easy_strerror(res);
            fprintf(stderr, "connect baidu server failed: %s\n",
                    curl_easy_strerror(res));
            engine->resultCode_ = (int)ai_engine::ErrorCode::NetError;
            return false;
        }
        engine->resultCode_ = (int)ai_engine::ErrorCode::Success;
        return true;
    }
    engine->resultCode_ = (int)ai_engine::ErrorCode::RequestFailed;
    return false;
}

void onWebSocket(bool (*doReceiveData)(BaiduSpeechEnginePrivate *),
                 BaiduSpeechEnginePrivate *engine) {
    while (true) {
        if (doReceiveData(engine)) {
            usleep(1000);
            continue;
        }
        return;
    }
}

void closeWebsocket(CURL *curl) {
    size_t sent;
    Json::Value bodyJson;
    bodyJson["type"] = "CANCEL";
    curl_ws_send(curl, bodyJson.toStyledString().c_str(),
                 bodyJson.toStyledString().size(), &sent, 0, CURLWS_BINARY);
}

bool writeOnceRecognitionResult(std::string data, intptr_t userData) {
    BaiduSpeechEnginePrivate *engine = (BaiduSpeechEnginePrivate *)userData;
    ai_engine::lm::speech::RecognitionResultCallback callback =
        engine->recognizingCallback_;
    ai_engine::lm::speech::RecognitionResultCallback finishedcallback =
        engine->recognizedCallback_;

    Json::Value resultJson = util::formatJsonFromString(data);
    if (token::isBaiduTokenExpiredByResult(resultJson)) {
        engine->needRefreshAccessToken_ = true;
        return false;
    }

    if (int errorCode = server_error::parseErrorCode(data)) {
        engine->lastErrorMessage_ = data;
        engine->resultCode_ = errorCode;
        fprintf(stderr, "Get baidu recognize result failed %s.\n",
                data.c_str());
        return false;
    }

    ai_engine::lm::speech::RecognitionResult result;
    engine->resultCode_ = (int)ai_engine::ErrorCode::Success;
    result.errorCode = (int)ai_engine::ErrorCode::Success;
    result.text = resultJson["result"][0].asString();
    result.type = ai_engine::lm::speech::ResultType::OnceRecognitionResult;
    callback(result);
    finishedcallback(result);
    return true;
}

bool doSendContinuousRecognitionSourceData(BaiduSpeechEnginePrivate *engine,
                                           int type, const char *data,
                                           int dataLength) {
    size_t sent;
    CURLcode resultCode =
        curl_ws_send(engine->currentCurl_, data, dataLength, &sent, 0, type);
    if (resultCode != CURLE_OK) {
        engine->lastErrorMessage_ = curl_easy_strerror(resultCode);
        fprintf(stderr, "send recognize source data to baidu failed: %s\n",
                curl_easy_strerror(resultCode));
        engine->resultCode_ = (int)ai_engine::ErrorCode::NetError;
        return false;
    }
    engine->resultCode_ = (int)ai_engine::ErrorCode::Success;
    return true;
}

bool doReceiveContinuousRecognitionResultData(
    BaiduSpeechEnginePrivate *engine) {
    size_t rlen;
    const curl_ws_frame *meta;
    char buffer[4100]{0};

    if (engine->isStopped_) {
        return false;
    }

    CURLcode curlCode = curl_ws_recv(engine->currentCurl_, buffer,
                                     sizeof(buffer), &rlen, &meta);
    if (curlCode == CURLE_AGAIN || curlCode == CURLE_RECV_ERROR) {
        return true;
    }
    if (curlCode != CURLE_OK) {
        engine->lastErrorMessage_ = curl_easy_strerror(curlCode);
        engine->resultCode_ = (int)ai_engine::ErrorCode::NetError;
        fprintf(stderr, "receive baidu recognize data failed: %s\n",
                engine->lastErrorMessage_.c_str());
        return false;
    }

    Json::Value recognitionJson =
        util::formatJsonFromString(std::string(buffer));
    if (recognitionJson["type"] == "HEARTBEAT") {
        return true;
    }
    if (int errorCode = server_error::parseErrorCode(std::string(buffer))) {
        if (errorCode == -3005 || errorCode == -3003) {
            return true;
        }
        engine->lastErrorMessage_ = std::string(buffer);
        engine->resultCode_ = errorCode;
        fprintf(stderr, "Get baidu recognize result failed %s.\n",
                std::string(buffer).c_str());
        return false;
    }

    ai_engine::lm::speech::RecognitionResult result;
    engine->resultCode_ = (int)ai_engine::ErrorCode::Success;
    result.errorCode = (int)ai_engine::ErrorCode::Success;
    result.text = recognitionJson["result"].asString();
    result.type =
        ai_engine::lm::speech::ResultType::ContinuousRecognitionResult;
    if (recognitionJson["type"].asString() == "FIN_TEXT") {
        engine->recognizedCallback_(result);
    } else {
        engine->recognizingCallback_(result);
    }
    return true;
}

bool writeOnceSynthesisResult(std::string data, intptr_t userData) {
    BaiduSpeechEnginePrivate *engine = (BaiduSpeechEnginePrivate *)userData;
    ai_engine::lm::speech::SynthesisResultCallback callback =
        engine->synthesizingCallback_;

    Json::Value resultJson = util::formatJsonFromString(data);
    if (token::isBaiduTokenExpiredByResult(resultJson)) {
        engine->needRefreshAccessToken_ = true;
        return false;
    }

    if (int errorCode = server_error::parseErrorCode(data)) {
        engine->lastErrorMessage_ = data;
        engine->resultCode_ = errorCode;
        fprintf(stderr, "Get baidu synthesize result failed %s.\n",
                data.c_str());
        return false;
    }

    ai_engine::lm::speech::SynthesisResult result;
    engine->resultCode_ = (int)ai_engine::ErrorCode::Success;
    result.audioFormat = "mp3";
    result.audioRate = 16000;
    result.audioChannel = 1;
    result.audioData = std::vector<char>{data.begin(), data.end()};
    result.errorCode = (int)ai_engine::ErrorCode::Success;
    result.type = ai_engine::lm::speech::ResultType::OnceSynthesisResult;
    callback(result);
    return true;
}

void doReceiveContinuousSynthesisResultData(BaiduSpeechEnginePrivate *engine,
                                            cpr::Url url, cpr::Header header,
                                            cpr::Body body) {
    cpr::Response response = cpr::Post(url, header, body);
    Json::Value resultJson = util::formatJsonFromString(response.text);
    if (engine->isStopped_) {
        return;
    }

    if (response.error) {
        engine->lastErrorMessage_ = response.error.message;
        engine->resultCode_ = (int)ai_engine::ErrorCode::NetError;
        fprintf(stderr, "Net error: %s.\n", response.error.message.c_str());
        return;
    }

    if (token::isBaiduTokenExpiredByResult(resultJson)) {
        engine->generateAccessToken_();
        return doReceiveContinuousSynthesisResultData(engine, url, header,
                                                      body);
    }

    if (int errorCode = server_error::parseErrorCode(response.text)) {
        engine->lastErrorMessage_ = response.text;
        engine->resultCode_ = errorCode;
        fprintf(stderr, "Get baidu synthesize result failed %s.\n",
                response.text.c_str());
        return;
    }

    ai_engine::lm::speech::SynthesisResult result;
    engine->resultCode_ = (int)ai_engine::ErrorCode::Success;
    result.audioFormat = "mp3";
    result.audioRate = 16000;
    result.audioChannel = 1;
    result.audioData =
        std::vector<char>{response.text.begin(), response.text.end()};
    result.errorCode = (int)ai_engine::ErrorCode::Success;
    result.type = ai_engine::lm::speech::ResultType::ContinuousSynthesisResult;
    engine->synthesizingCallback_(result);
}

BaiduSpeechEnginePrivate::BaiduSpeechEnginePrivate() {}

BaiduSpeechEnginePrivate::~BaiduSpeechEnginePrivate() { releaseWebResources(); }

std::string BaiduSpeechEnginePrivate::backendName() {
    return std::string("baidu");
}

void BaiduSpeechEnginePrivate::setAppId(const std::string &appId) {
    appId_ = appId;
}

void BaiduSpeechEnginePrivate::setApiKey(const std::string &apiKey) {
    apiKey_ = apiKey;
}

void BaiduSpeechEnginePrivate::setSecretKey(const std::string &secretKey) {
    secretKey_ = secretKey;
}

void BaiduSpeechEnginePrivate::setRecognizingCallback(
    ai_engine::lm::speech::RecognitionResultCallback callback) {
    recognizingCallback_ = callback;
}

void BaiduSpeechEnginePrivate::setRecognizedCallback(
    ai_engine::lm::speech::RecognitionResultCallback callback) {
    recognizedCallback_ = callback;
}

void BaiduSpeechEnginePrivate::setSynthesizingCallback(
    ai_engine::lm::speech::SynthesisResultCallback callback) {
    synthesizingCallback_ = callback;
}

void BaiduSpeechEnginePrivate::setSynthesizedCallback(
    ai_engine::lm::speech::FinishedCallback callback) {
    synthesizedCallback_ = callback;
}

ai_engine::ErrorCode
BaiduSpeechEnginePrivate::initContinuousRecognitionModule() {
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::startContinuousRecognition(
    const std::string &params) {
    if (!recognizingCallback_ || !recognizedCallback_) {
        return ai_engine::ErrorCode::ComponentMissing;
    }

    isStopped_ = false;
    if (!initWebResources("recognize")) {
        return (ai_engine::ErrorCode)resultCode_;
    }

    setRecognitionParams_(params);

    Json::Value bodyJson;
    bodyJson["data"]["appid"] = std::stoi(appId_);
    bodyJson["data"]["appkey"] = apiKey_;
    bodyJson["data"]["dev_pid"] = 15372;
    bodyJson["data"]["cuid"] = "kylin-ai-subsystem";
    bodyJson["data"]["format"] = "pcm";
    bodyJson["data"]["sample"] = 16000;
    bodyJson["type"] = "START";
    doSendContinuousRecognitionSourceData(this, CURLWS_TEXT,
                                          bodyJson.toStyledString().c_str(),
                                          bodyJson.toStyledString().length());
    return (ai_engine::ErrorCode)resultCode_;
}

ai_engine::ErrorCode
BaiduSpeechEnginePrivate::writeContinuousRecognitionAudioData(
    const std::vector<char> &audioData) {
    if (audioData.empty()) {
        return ai_engine::ErrorCode::ParamError;
    }
    if (!recognizingCallback_ || !recognizedCallback_) {
        return ai_engine::ErrorCode::ComponentMissing;
    }

    if (isStopped_) {
        return (ai_engine::ErrorCode)resultCode_;
    }

    if (!currentCurl_) {
        return resultCode_ == (int)ai_engine::ErrorCode::Success
                   ? ai_engine::ErrorCode::ComponentMissing
                   : (ai_engine::ErrorCode)resultCode_;
    }

    doSendContinuousRecognitionSourceData(this, CURLWS_BINARY, audioData.data(),
                                          audioData.size());
    return (ai_engine::ErrorCode)resultCode_;
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::stopContinuousRecognition() {
    releaseWebResources();
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode
BaiduSpeechEnginePrivate::destroyContinuousRecognitionModule() {
    releaseWebResources();
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::initRecognizeOnceModule() {
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode
BaiduSpeechEnginePrivate::recognizeOnce(const std::string &params,
                                        const std::vector<char> &audioData) {
    if (audioData.empty()) {
        return ai_engine::ErrorCode::ParamError;
    }
    if (!recognizingCallback_ || !recognizedCallback_) {
        return ai_engine::ErrorCode::ComponentMissing;
    }

    setRecognitionParams_(params);

    Json::Value postJson;
    postJson["format"] = format_;
    postJson["rate"] = 16000;
    postJson["channel"] = 1;
    postJson["cuid"] = "kylin-ai-subsystem";
    postJson["token"] = getAccessToken_();
    postJson["dev_pid"] = 80001;
    postJson["speech"] =
        util::base64Encode(std::string(audioData.data(), audioData.size()));
    postJson["len"] = (int)audioData.size();
    cpr::Url url{cpr::Url{"https://vop.baidu.com/pro_api"}};
    cpr::Header header{cpr::Header{{"Content-Type", "application/json"}}};
    cpr::Body body{cpr::Body{{postJson.toStyledString()}}};
    cpr::Response response = cpr::Post(
        url, header, body,
        cpr::WriteCallback{writeOnceRecognitionResult, (intptr_t)this});

    if (!processCprResponse_(response, "recognize")) {
        return recognizeOnce(params, audioData);
    } else {
        return (ai_engine::ErrorCode(resultCode_));
    }
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::destroyRecognizeOnceModule() {
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::initContinuousSynthesisModule() {
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode
BaiduSpeechEnginePrivate::startContinuousSynthesis(const std::string &params) {
    if (!synthesizingCallback_ || !synthesizedCallback_) {
        return ai_engine::ErrorCode::ComponentMissing;
    }

    isStopped_ = false;
    setSynthesisParams_(params);
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::writeContinuousSynthesisText(
    const std::string &text) {
    if (text.empty()) {
        return ai_engine::ErrorCode::ParamError;
    }
    if (!synthesizingCallback_ || !synthesizedCallback_) {
        return ai_engine::ErrorCode::ComponentMissing;
    }

    if (isStopped_) {
        return (ai_engine::ErrorCode)resultCode_;
    }

    if (!initWebResources("Synthesize", text)) {
        return (ai_engine::ErrorCode)resultCode_;
    }

    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::stopContinuousSynthesis() {
    releaseWebResources();
    if (!synthesizingCallback_ || !synthesizedCallback_) {
        return ai_engine::ErrorCode::ComponentMissing;
    }
    synthesizedCallback_(
        ai_engine::lm::speech::ResultType::ContinuousSynthesisResult);
    return ai_engine::ErrorCode(resultCode_);
}

ai_engine::ErrorCode
BaiduSpeechEnginePrivate::destroyContinuousSynthesisModule() {
    releaseWebResources();
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::initSynthesizeOnceModule() {
    return ai_engine::ErrorCode::Success;
}

ai_engine::ErrorCode
BaiduSpeechEnginePrivate::synthesizeOnce(const std::string &params,
                                         const std::string &text) {
    if (text.empty()) {
        return ai_engine::ErrorCode::ParamError;
    }
    if (!synthesizingCallback_ || !synthesizedCallback_) {
        return ai_engine::ErrorCode::ComponentMissing;
    }

    setSynthesisParams_(params);

    std::string postString;
    postString += ("tex=" + util::urlEncode(util::urlEncode(text)));
    postString += ("&tok=" + getAccessToken_());
    postString += ("&cuid=kylin-ai-subsystem");
    postString += ("&ctp=1");
    postString += ("&lan=zh");
    postString += ("&spd=" + std::to_string(speed_));
    postString += ("&pit=" + std::to_string(pitch_));
    postString += ("&vol=" + std::to_string(volume_));
    postString += ("&per=1");
    postString += ("&aue=3");
    cpr::Url url{cpr::Url{"https://tsn.baidu.com/text2audio"}};
    cpr::Header header{
        cpr::Header{{"accept", "*/*"},
                    {"Content-Type", "application/x-www-form-urlencoded"}}};
    cpr::Body body{cpr::Body{{postString}}};
    cpr::Response response =
        cpr::Post(url, header, body,
                  cpr::WriteCallback{writeOnceSynthesisResult, (intptr_t)this});

    if (!processCprResponse_(response, "synthesize")) {
        return synthesizeOnce(params, text);
    } else {
        synthesizedCallback_(
            ai_engine::lm::speech::ResultType::OnceSynthesisResult);
        return (ai_engine::ErrorCode(resultCode_));
    }
}

ai_engine::ErrorCode BaiduSpeechEnginePrivate::destroySynthesizeOnceModule() {
    return ai_engine::ErrorCode::Success;
}

void BaiduSpeechEnginePrivate::setRecognitionParams_(
    const std::string &params) {
    Json::Value paramsJson = util::formatJsonFromString(params);
    format_ = paramsJson["format"].asString();
    rate_ = paramsJson["rate"].asInt();
    channel_ = paramsJson["channel"].asInt();
    if (format_.empty()) {
        format_ = "pcm";
    }
    if (rate_ == 0) {
        rate_ = 16000;
    }
    if (channel_ == 0) {
        channel_ = 1;
    }
}

void BaiduSpeechEnginePrivate::setSynthesisParams_(const std::string &params) {
    Json::Value paramsJson = util::formatJsonFromString(params);
    speed_ = paramsJson["speed"].asInt();
    volume_ = paramsJson["volume"].asInt();
    pitch_ = paramsJson["pitch"].asInt();
    if (speed_ == 0) {
        speed_ = 5;
    }
    if (volume_ == 0) {
        volume_ = 5;
    }
    if (pitch_ == 0) {
        pitch_ = 5;
    }
}

std::string &BaiduSpeechEnginePrivate::getAccessToken_() {
    if (accessToken_.empty()) {
        generateAccessToken_();
    }
    return accessToken_;
}

void BaiduSpeechEnginePrivate::generateAccessToken_() {
    bool hasNetError = false;
    accessToken_ = token::getBaiduSpeechToken(apiKey_, secretKey_, hasNetError,
                                              lastErrorMessage_);
    if (accessToken_.empty()) {
        hasNetError
            ? resultCode_ = (int)ai_engine::ErrorCode::NetError
            : resultCode_ = (int)ai_engine::ErrorCode::AuthenticationFailed;
        fprintf(stderr, "Get baidu speech token failed.\n");
        return;
    }
    resultCode_ = (int)ai_engine::ErrorCode::Success;
}

bool BaiduSpeechEnginePrivate::initWebResources(std::string type,
                                                const std::string &params) {
    if (currentThread_) {
        delete currentThread_;
        currentThread_ = nullptr;
    }

    currentCurl_ = curl_easy_init();
    if (!connectWebSocket(currentCurl_, this)) {
        if (currentCurl_) {
            curl_easy_cleanup(currentCurl_);
            currentCurl_ = nullptr;
        }
        return false;
    }

    if (type == "recognize") {
        currentThread_ = new std::thread(
            onWebSocket, doReceiveContinuousRecognitionResultData, this);
        currentThread_->detach();
    } else {
        std::string postString;
        postString += ("tex=" + util::urlEncode(util::urlEncode(params)));
        postString += ("&tok=" + getAccessToken_());
        postString += ("&cuid=kylin-ai-subsystem");
        postString += ("&ctp=1");
        postString += ("&lan=zh");
        postString += ("&spd=" + std::to_string(speed_));
        postString += ("&pit=" + std::to_string(pitch_));
        postString += ("&vol=" + std::to_string(volume_));
        postString += ("&per=1");
        postString += ("&aue=3");
        cpr::Url url{cpr::Url{"https://tsn.baidu.com/text2audio"}};
        cpr::Header header{
            cpr::Header{{"accept", "*/*"},
                        {"Content-Type", "application/x-www-form-urlencoded"}}};
        cpr::Body body{cpr::Body{{postString}}};

        currentThread_ = new std::thread(doReceiveContinuousSynthesisResultData,
                                         this, url, header, body);
        currentThread_->detach();
    }
    return true;
}

void BaiduSpeechEnginePrivate::releaseWebResources() {
    isStopped_ = true;
    if (currentCurl_) {
        closeWebsocket(currentCurl_);
        curl_easy_cleanup(currentCurl_);
        currentCurl_ = nullptr;
    }
    if (currentThread_) {
        delete currentThread_;
        currentThread_ = nullptr;
    }
}

bool BaiduSpeechEnginePrivate::processCprResponse_(cpr::Response &response,
                                                   std::string type) {
    if (needRefreshAccessToken_) {
        generateAccessToken_();
        needRefreshAccessToken_ = false;
        return false;
    }

    if (response.error.code == cpr::ErrorCode::REQUEST_CANCELLED &&
        resultCode_ > 0) {
        if (type == "synthesize") {
            resultCode_ =
                (int)server_error::errorCode2speechResult(resultCode_);
        } else {
            resultCode_ =
                (int)server_error::errorCode2speechResult(resultCode_);
        }
        return true;
    }

    if (response.error) {
        lastErrorMessage_ = response.error.message;
        resultCode_ = (int)ai_engine::ErrorCode::NetError;
        fprintf(stderr, "Net error: %s.\n", response.error.message.c_str());
        return true;
    }

    resultCode_ = (int)ai_engine::ErrorCode::Success;
    return true;
}
